(* Mathematica Package *)
(* Created by the Wolfram Workbench 14.04.2016 *)

(* :Title:  	Quantum Algebra *)

(* :Context:  	QuantumAlgebra` *)

(* :Author:  	Juergen Habelt
				
				Based on the work of:
				
				Cesar Guerra G.
             	guerra_cesar@yahoo.com
             
             	Pontificia Universidad Catolica del Peru 
             	Av. Universitaria cdra. 18, San Miguel
             	Lima-32 PERU. Telf. (511) 4602870 *)

(* :Summary:  	Quantum Algebra is a package to perform Quantum Calculations
              	using non commutative algebra. Dirac Notation is implemented. *)

(* :Mathematica Version: 
				
				tested with Mathematica v10.2 
				may work with other versions too   *)

(*	Changes to old source
	- package works with new versions of Mathematica
	- notations, alias definitions, styles are defined in Module instead of a Notebook
	- alias definitions added for operators with Sub-/Superscripts
	- introduced a new type of Constant: ComplexConstantQ 
	  (old lives forth as SimpleConstantQ)
	- new rules for the ComplexConstantQ type enable further simplification
	- stricter definition of KetQ, BraQ
	- stricter definition of the SimpleConstantQ / ComplexConstantQ / ConstantQ types
	- stricter definition of OperatorQ
	- encapsulation of the strict definitions in the Analyze module
	- type usage can be controlled by an option, so one can switch back to old definition usage
	- options in conjunction with QAPrint allow for controlled output of diagnostic messages
	- code formatted, comments added, sections added
  	- added a rule in the QASet section, support for differential operators 
  	  (like Bra[] ** Operator[] ** Ket[] = D[BraKet[], ...])
  	- added test cases for the most of the functionality
  	- added documentation as regular Mathematica application documentation
	- bug fixes
	  -- usage of the Operator[1] in the Exp section
	  -- changed variable names in the BCHRelation section (undefined variables)
	  -- fixed all code errors / warnings
	  -- changed the order of QASet rules, because some rules have not been recognized
	  -- one QASet rule contained mixed usage of Set / SetDelayed, probably forgotten to change to SetDelayed
*)

(*	TODO: Ideas
	- create a block structure (similar to Begin - End), which can be used to express assumptions
	  -- assumptions for Simplify etc.
	  -- additional assumptions for QA (TBD)
	- stricter rules
	- generate a Commutator type
	  -- possible advantage: some rules permit a more general call syntax (_Commutator -> _?CommutatorQ)
	  -- Attention! Be careful in design of rule precedence (QASet)! 
	  
	TODO: OPEN ISSUES
	- complete test cases, goal high code coverage
	- complete documentation, 
	  -- more usage scenarios
	  -- better instructions, how to use QA features
*)

BeginPackage[
	"QuantumAlgebra`",
	{
		"QuantumAlgebra`Derivation`",
		"QuantumAlgebra`Analyze`",
		"QuantumAlgebra`Options`",
		"QuantumAlgebra`Styles`",
		"QuantumAlgebra`QuantumNotation`",
		"Notation`"
	}
]

(* ::Section:: *)
(* Region Exported Symbols *)


Operator::usage =
	"Operator[A__] defines an operator object, where sequence A labels the operator. " <>
	"Some other forms are defined as follows. \n" <>
	"Operator[A,{x__},{}] defines an operator where x is a sequence of subscript indices. \n" <> 
	"Operator[A,{},{y__}] defines an operator where y is a sequence of superscript indices. \n" <> 
	"Operator[A,{x__},{y__}] defines an operator where both subscripts and superscrips are given. " <>
	"If y is an integer this is interpreted as Power. \n" <>
	"Operator[__][t__] contains additional variables. \n" <>
	"Operator[A__,\"H\",__] contains a label \"H\" signifying a hermitian operator for " <>
	"Operator[A]. \n" <>
	"Operator[1] and Operator[0] are interpreted as the identity and zero operators respectively.\n\n" <>
	"ALIASES: \n" <>
	"[ESC] op [ESC]  : operator. \n" <>
	"[ESC] hop [ESC] : hermitian operator."


Ket::usage =
	"Ket[A__] is the simple form of a Ket object where the sequence A labels the Ket. " <>
	"Additional forms allow the usage of indices. \n" <>
	"Ket[A__,{x__},{}] enables the use of a sequence of subscript indices (x). \n" <>
	"Ket[A__,{},{y__}] provides a sequence of superscripts (y). \n" <>
	"Ket[A__,{x__},{y__}] provides subscripts as well as superscripts. \n\n" <>
	"ALIASES: \n" <>
	"[ESC] ket [ESC]   : ket\n" <>
	"[ESC] ketd [ESC]  : ket with subscript\n" <>
	"[ESC] ketu [ESC]  : ket with superscript\n" <>
	"[ESC] ketdu [ESC] : ket with both scripts.\n\n" <>
	"Ket is also used as a tag for TagBoxes that guide the " <>
	"interpretation of the internal structure of ket notation."


Bra::usage =
	"Bra[A__] is the simple form of a Bra object where the sequence A labels the Bra. " <>
	"Additional forms allow the usage of indices. \n" <>
	"Bra[A__,{x__},{}] enables the use of a sequence of subscript indices (x). \n" <>
	"Bra[A__,{},{y__}] provides a sequence of superscripts (y). \n" <>
	"Bra[A__,{x__},{y__}] provides subscripts as well as superscripts. \n\n" <>
	"ALIASES: \n" <>
	"[ESC] bra [ESC]   : bra\n" <>
	"[ESC] dbra [ESC]  : bra with subscript\n" <>
	"[ESC] ubra [ESC]  : bra with superscript\n" <>
	"[ESC] dubra [ESC] : bra with both scripts.\n\n" <>
	"Bra is also used as a tag for TagBoxes that guide the " <>
	"interpretation of the internal structure of bra notation."

BraKet::usage =
	"BraKet[{A__},{B__}] is the simple form of a BraKet object where the sequences " <> 
	"A and B label the BraKet. Additional forms allow the usage of indices. " <>
	"Whereas the indices of Kets are on their right-hand side and the indices " <>
	"of Bras are on their left-hand side, BraKets can have their indices placed on " <>
	"both sides. \n" <>
	"BraKet[{A__,{x__},{}},{B__}] has placed subscripts (x) on the left-hand side. \n" <>
	"BraKet[{A__,{},{y__}},{B__}] has superscripts (y) on the left. \n" <>
	"BraKet[{A__,{x__},{y__}},{B__,{u__},{v__}}] has all kinds of indices. " <>
	"More combinations are posible. \n\n" <>
	"ALIASES: \n" <>
	"[ESC] braket [ESC]     : braket\n" <>
	"[ESC] braketd [ESC]    : braket with right-handed subscript\n" <>
	"[ESC] ubraket [ESC]    : braket with left-handed superscript\n" <>
	"[ESC] dubraketdu [ESC] : braket with both scripts on both sides. \n" <>
	"More combinations are possible using u(up) and d(down). \n\n" <>
	"BraKet is also used as a tag for TagBoxes that guide the " <>
	"interpretation of the internal structure of braket notation."

OperatorQ::usage =
	"OperatorQ[ex] checks if the expression ex is an Operator."

KetQ::usage =
	"KetQ[ex] checks if the expression ex is a Ket."

BraQ::usage =
	"BraQ[ex] checks if the expression ex is a Bra."

ConstantQ::usage =
	"ConstantQ[ex] checks if the expression ex is a Constant. " <>
	"The improved Quantum Algebra application supports simple as well as " <>
	"complex constants. Whereas simple constants do not contain any quantum " <>
	"object, complex constants consist of sequences like (Bra ** Operator ** Ket) " <>
	"which are known to evaluate to numbers. This allows for the simplification of " <>
	"expressions."

Commutator::usage =
	"Commutator[A,B] represents the commutator between operators A and B. \n\n" <>
	"ALIASES: \n" <>
	"[ESC] com [ESC] : commutator."

Hermitian::usage = 
	"Hermitian[expr] performs the hermitian operation on expr. " <>
	"An expression with a dagger as a superscript is interpreted " <>
	"as the hermitian operation on expression."

QAPowerExpand::usage =
	"QAPowerExpand[expr] expands all powers in expr as non-commutative products."

QAPowerContract::usage =
	"QAPowerContract[expr] is intented to contract terms like: " <>
	"A^n ** A^m to A^(n+m) where A is an operator and n and m simple exponents."

QAProductExpand::usage =
	"QAProductExpand[expr] expands all non-commutative products of " <>
	"constant terms and quantum objects in expr as simple products " <>
	"of constant terms and non commutative-products of quantum objects."

QACommutatorExpand::usage =
	"QACommutatorExpand[expr] expands all commutators in expr."

QACommutatorContract::usage =
	"QACommutatorContract[expr] contracts all expanded commutators in expr. \n" <>
	"QACommutatorContract[expr,{A,B}] contracts all expanded forms of Commutator[A,B] " <> 
	"in expr. \n" <>
	"QACommutatorContract[expr,{{A,B},...}] performs several commutators' contracts."

QAExpandAll::usage =
	"QAExpandAll[expr] performs the following commands in expr: " <>
	"QAPowerExpand, QAProductExpand, and QACommutatorExpand in that order. "

QASeries::usage =
	"QASeries[expr,{A,B,n}] expands the operator expr depending " <>
	"only on A, in power series of A around the operator B up to " <>
	"order n. The operator B must be of the form c Operator[1] or " <>
	"Operator[0] (zero). \n" <>
	"QASeries[Exp[A] ** B ** Exp[-A],n] expands the first argument " <>
	"as a series expansion in terms of operators A and B up to n " <>
	"terms. \n" <>
	"QASeries does not return a series object, but a polynomial " <>
	"on operator A."

QAOrdering::usage =
	"QAOrdering[expr_, list_] transforms expr, which is formed by " <>
	"a product of operators to an equivalent expression in which the " <>
	"order of the products is changed. The new order will be as " <>
	"given in list from left to right."

BCHRelation::usage =
	"BCHRelation[expr] uses the Baker Campbell Hausdorff relation " <>
	"to expand all operators of the form Exp[A+B] in expr where " <>
	"A and B are operators. BCHRelation automatically checks if " <>
	"A and B commute with their commutator."

QASet::usage =
	"QASet[expr1 = expr2] and QASet[expr1 := expr2] where expr1 and " <>
	"expr2 involve quantum objects like operators, bras, kets, ... , " <>
	"first evaluates the Set and SetDelayed operations as given in " <>
	"its argument. After that, QASet tries to set up definitions that " <>
	"can be deduced from the first relation. \n" <>
	"For example if we evaluate QASet[ Commutator[A,B] = C ] then " <>
	"the argument is evaluated in the usual way. But from this set " <>
	"it can be deduced that Commutator[B,A] = -C, so QASet also " <>
	"does this assignment. \n" <>
	"QASet can handle the following definitions: \n" <>
	" - QASet[ Com[A,B] = expr ] \n" <>
	" - QASet[ Com[A,Com[A,B]] = expr ] \n" <>
	" - QASet[ Com[Com[A,B],Com[C,D]] = expr ] \n" <>
	" - QASet[ Com[A,Com[A,B]] = 0, Com[B,Com[A,B]] = 0 ] \n" <>
	" - QASet[ A ** Ket[n] = a Ket[n] ] \n" <>
	"where A, B, C, D are operators and Com means Commutator. Also " <>
	":= can be used  instead of = if neccesary. For a full dicussion " <>
	"see QuantumAlgebra.nb Section 8. "

QAClear::usage =
	"QAClear[ lhs =. ] removes all definitions made by QASet associated " <>
	"with the definition of lhs. See QASet. "

SuperDagger::usage =
	"SuperDagger[ex] performs the same operation as Hermitian[ex]."

SuperStar::usage =
	"SuperStar[ex] performs the same operation as Conjugate[ex]."

General::invalid =
	"Warning: The operator Times cannot be used to concatenate two Quantum objects:\n" <>
	"`1` * `2`\n" <>
	"Attention! NonCommutativeMultiply precedes Times in operator precedence.\n" <>
	"Trying to recover using NonCommutativeMultiply instead of Times."


(* ::Section:: *)
(* Region Implementation of the Package *)

Begin["`Private`"]


protected = Unprotect[NonCommutativeMultiply, Exp, Times, Power]


(* ::Subsection:: *)
(* Region DEFINING THE OBJECTS *)


SetAttributes[{OperatorQ, KetQ, BraQ, ConstantQ, QAProductExpand,
	QAPowerExpand, QACommutatorExpand, QAExpandAll}, Listable]

 
ComplexConstantQ[expr_NonCommutativeMultiply] /;
	OptionValue[QAOptions, DefinitionLevel] == 2 &&
	Length[expr] >= 3  :=
	Module[{},
		If[MatchQ[List @@ expr, { _Bra, _?OperatorQ.., _Ket }],
			Return[True];
		];
		False
	]

ComplexConstantQ[expr_] /;
	OptionValue[QAOptions, DefinitionLevel] == 3 &&
	QuantumConstantExpressionQ[expr] :=
	True
	
ComplexConstantQ[expr_] := False 


SimpleConstantQ[expr_, mrk_String: "def"] /;
	OptionValue[QAOptions, DefinitionLevel] <= 2 :=
	(QuantumAlgebra`Options`Private`QAPrint["SimpleConstantQ: ", mrk, ": ", expr];
	FreeQ[expr, Operator] && FreeQ[expr, Ket] && FreeQ[expr, Bra])

SimpleConstantQ[expr_, mrk_String: "def"] /;
	OptionValue[QAOptions, DefinitionLevel] >= 3 :=
	SimpleConstantExpressionQ[expr]


ConstantQ[expr_] /;
	SimpleConstantQ[expr, "ConstantQ"] || ComplexConstantQ[expr] :=
	(ConstantCache[expr];
	True)
	
ConstantQ[expr_] := False

 
ConstantCache[inp_: Null] := 
	Module[{val},
		If[inp === Null,
			val = cache;
			Return[val // ReleaseHold]];
		cache = Hold[inp];
		True
	]
 
cache = 0


(*
	Improved OperatorQ Predicate
	- can be controlled by DefinitionLevel flag
	- 1: take old definitions
	- 2: take new definitions (OperatorQv2)
	- 3: take strict definitions
*)
OperatorQ[t_] /;
	OptionValue[QAOptions, DefinitionLevel] == 1 := 
	!FreeQ[t, Operator]      

OperatorQ[t_] /;
	OptionValue[QAOptions, DefinitionLevel] == 2 := 
	OperatorQv2[t]

OperatorQ[t_] /;
	OptionValue[QAOptions, DefinitionLevel] == 3 := 
	OperatorExpressionQ[t]      


KetQ[t_] := !FreeQ[t, Ket, 1]
 
BraQ[t_] := !FreeQ[t, Bra, 1]


(* ::Subsection:: *)
(* Region Level 2 Operator Definition *)

OperatorQWv2[expr_Operator] :=
	True

OperatorQWv2[expr_] /; 
	FreeQ[expr, Operator] :=
	False

OperatorQWv2[expr_Plus] /; 
	MatchQ[List @@ expr, { ___, _?OperatorQv2, ___ }] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] := 
	True
	
OperatorQWv2[expr_NonCommutativeMultiply] /; 
	MatchQ[List @@ expr, { ___, _?OperatorQv2, ___ }] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] := 
	True

OperatorQWv2[expr_Times] /; 
	MatchQ[List @@ expr, { _?(SimpleConstantQ[#, "20."]&)..., _?OperatorQv2, _?(SimpleConstantQ[#, "21."]&)... }] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] := 
	True

OperatorQWv2[expr_Power] /; 
	MatchQ[List @@ expr, { _?OperatorQv2, _Integer?NonNegative }] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] := 
	True

OperatorQWv2[expr_Power] /; 
	MatchQ[List @@ expr, { E, _?OperatorQv2 }] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] :=
	True

OperatorQWv2[expr_Commutator] /;
	Length[expr] == 2 &&
	OperatorQv2[First[expr]] &&
	OperatorQv2[Last[expr]] :=
	True

OperatorQWv2[expr_] /; 
	Length[expr] >= 1 &&
	ContainsAny[Head[expr], KnownFunctions[]] &&
	MatchQ[First[expr], _?OperatorQv2] &&
	FreeQ[expr, Bra] &&
	FreeQ[expr, Ket] :=
	True

OperatorQWv2[expr_] :=
	Module[{ nested = Nested[expr, KnownFunctions[]] },
		QuantumAlgebra`Options`Private`QAPrint[1, "Last resort test for Operator."];
		Which[
			nested === Null,
			QuantumAlgebra`Options`Private`QAPrint[1, "Last resort test for Operator failed."],
			OperatorQv2[nested],
			QuantumAlgebra`Options`Private`QAPrint[1, "Last resort test for Operator succeeded: ", nested],
			True,
			QuantumAlgebra`Options`Private`QAPrint[1, "Last resort test for Operator failed: ", nested]
		]
	]
	
OperatorQv2 := QuantumAlgebra`Options`Private`LevelUpdater[OperatorQWv2]

Nested[expr_, funcs_List] :=
	Module[{res, kf = KnownFunctions[]},
		res = NestWhile[
			First, 
			expr, 
			(!ContainsAny[Head[#], kf] && Length[#] > 0)&
		];
		If[res === expr || !ContainsAny[res, kf],
			Null,
			res
		]
	]

KnownFunctions[] :=
	{
		Sin, 
		Cos, 
		Log, 
		Simplify, 
		FullSimplify, 
		Expand, 
		PowerExpand, 
		QAProductExpand
	}


(* ::Subsection:: *)
(* Region ALGEBRA FOR NON COMMUTATIVE MULTIPLY *)


 0 P_?OperatorQ := Operator[0]
 
 c_ Operator[0] /; ConstantQ[c] := Operator[0]

 0 ** P_ := 0
 
 P_ ** 0 := 0
 
 Operator[0] ** P_?OperatorQ := Operator[0]
 
 P_?OperatorQ ** Operator[0] := Operator[0]
 
 Operator[0] ** P_ := 0
 
 P_ ** Operator[0] := 0
 
 Operator/: Operator[0] + P_ := P
 
 Operator/: Operator[0] a_ := 0
 
 Operator[1] ** P_ := P
 
 P_ ** Operator[1] := P 
 
 P_ ** Q_Plus := Map[(P ** #)&, Q]

 Q_Plus ** P_ := Map[(# ** P)&, Q]

 (P_ ** c_ **  Q_) /; ConstantQ[c] := ConstantCache[] P ** Q

 (c_ **  P_ ** Q_) /; ConstantQ[c] := ConstantCache[] P ** Q

 (P_ ** (c_ Q_)) /; ConstantQ[c] := ConstantCache[] P ** Q

 ((c_ P_) ** Q_) /; ConstantQ[c] := ConstantCache[] P ** Q
 
 P_ ** Sum[c_ Q_, rng_] /; SimpleConstantQ[c] := Sum[c P ** Q, rng]

 Sum[c_ P_, rng_] ** Q_ /; SimpleConstantQ[c] := Sum[c P ** Q, rng]


(* ::Subsection:: *)
(* Region THE INNER PRODUCT *)

 
 Bra[m__] ** Ket[n__] := BraKet[{m},{n}]

 BraKet[x__] ** BraKet[y__] := BraKet[x] BraKet[y]

 BraKet/: Conjugate[BraKet[{m__},{n__}]] := BraKet[{n},{m}]


(* ::Subsection:: *)
(* Region INTEGER SUPERSCRIPT INTERPRETATION AS POWER *)


 Power[Operator[1], n_?IntegerQ ] := Operator[1]

 Power[Operator[0], n_?IntegerQ ] := 0

 Operator[P_, {}, { n_?IntegerQ }] := Power[Operator[P], n]

 Power[Operator[P_], n_?IntegerQ][t__] := Power[Operator[P][t], n]

 Operator[P_, {x__}, { n_?IntegerQ }] := Power[Operator[P, {x}, {}], n]

 Power[Operator[P_, {x__}, {}], n_?IntegerQ][t__] := Power[Operator[P, {x}, {}][t], n]

 Power[expr_, 0] := Operator[1] /; OperatorQ[expr] && expr =!= Operator[0] 

 Power[Operator[0], 0] := Indeterminate

 Power/: P_^(-1) ** P_ := Operator[1] /; OperatorQ[P]

 Power/: P_ ** P_^(-1) := Operator[1] /; OperatorQ[P]


(* ::Subsection:: *)
(* Region RULES FOR EXPONENTIAL OPERATORS *)

(*	TODO:
	- instead of . Times is recognized in input pattern -> OKAY
	- shouldn't the first rule not contain an Operator[1] multiplied ..
	  so the Operator character is saved? -> done.
*)
 Power[E, Operator[0]] := Operator[1]

 Power[E, a_. Operator[0]] := Operator[1] /; SimpleConstantQ[a, "5."] 

 Power[E, a_. Operator[1]] := Exp[a] Operator[1] /; SimpleConstantQ[a, "5."] 

 (* OBSOLETE - erroneous
 	Power/: Power[E, P_] ** Power[E,d_.+ Q_] := 
 	Power[E,d] Power[E,P] ** Power[E,Q] /; 
 	OperatorQ[P] && OperatorQ[Q] && ConstantQ[c] && ConstantQ[d] 
 *)

 Power/: Power[E, P_] ** Power[E, Q_] := Operator[1] /; OperatorQ[P] && OperatorQ[Q] && P == -Q
 
 Power/: Power[E, c_. P_] ** Power[E, d_. P_] := Exp[(c + d) P] /; OperatorQ[P] 

 HoldPattern[Exp[a_. Operator[0]]] := Operator[1]


 (* TODO: check: instead of Q, P in com2 B, A has been used (NOT DEFINED)
 	TODO: a few Power patterns have been commented out, see above	 
 *)
 
 SetAttributes[BCHRelation, HoldAllComplete]
 
 BCHRelation[expr_] := 
 	Module[ 
 		{sol}, 
 		ClearAttributes[Plus, Orderless] ;
 		sol = expr /.  Power[E, c_. (P_ + Q_)]  :> 
		Module[
			{ com1 = c^2 Commutator[P, Q], com2 = c^2 Commutator[Q, P]},        
		    If[Commutator[c P, com1] === Commutator[c Q, com1] === 0 ||
		       Commutator[c P, com2] === Commutator[c Q, com2] === 0 ||
		       Commutator[c P, com2] === Commutator[c Q, com1] === 0 ||
		       Commutator[c P, com1] === Commutator[c Q, com2] === 0 ||              
               Commutator[c P, com1] === Commutator[com1, c Q] === 0 ||
		       Commutator[c P, com2] === Commutator[com2, c Q] === 0 ||
               Commutator[c P, com1] === Commutator[com2, c Q] === 0 ||
		       Commutator[c P, com2] === Commutator[com1, c Q] === 0 ||
               Commutator[com1, c P] === Commutator[c Q, com1] === 0 ||
		       Commutator[com1, c P] === Commutator[com1, c Q] === 0 ||
               Commutator[com1, c P] === Commutator[c Q, com2] === 0 ||
		       Commutator[com2, c P] === Commutator[com1, c Q] === 0 ||	     
		       Commutator[com2, c P] === Commutator[c Q, com2] === 0 ||
		       Commutator[com2, c P] === Commutator[com2, c Q] === 0 ||
               Commutator[com2, c P] === Commutator[c Q, com1] === 0 ||
		       Commutator[com1, c P] === Commutator[com2, c Q] === 0 ,
			   QAProductExpand[ Exp[c P] ** Exp[c Q] ** Exp[-com1/2]], 
			   Exp[c (P + Q)]
			]
		]  /; OperatorQ[P] && OperatorQ[Q] && ConstantQ[c] ;
        SetAttributes[Plus, Orderless];
		sol]


(* ::Subsection:: *)
(* Region RULES FOR HERMITIAN OPERATION *)


 SuperDagger = Hermitian;

 SuperStar = Conjugate;

 Hermitian[Operator[0]] := Operator[0]

 Hermitian[Operator[1]] := Operator[1]

 Hermitian[Operator[P_, x___]] := Operator[P, "H", x]

 Hermitian[Operator[P_, "H", x___]] := Operator[P, x]

 Hermitian[BraKet[{a__}, {b__}]] := BraKet[{b}, {a}]					(* added 2017-02-21 *)

 Hermitian[c_] := Conjugate[c] /; SimpleConstantQ[c]

 Hermitian[c_ P_] := 
 	Conjugate[c] Hermitian[P] /; 
 	SimpleConstantQ[c, "6."] && !SimpleConstantQ[P, "7."]

 Hermitian[P_ + Q_] := 
 	Hermitian[P] + Hermitian[Q] /; 
 	!SimpleConstantQ[P, "8."] && !SimpleConstantQ[Q, "9."]

 Hermitian[ P_ ** Q_ ] := Hermitian[Q] ** Hermitian[P]

 Hermitian[Ket[a__]] := Bra[a]

 Hermitian[Bra[a__]] := Ket[a]

 Hermitian[Operator[P__][t__]] := Hermitian[Operator[P]][t]

 Hermitian[Power[P_, n_]] := Power[Hermitian[P], n] /; OperatorQ[P]

 Hermitian[Power[E, P_]] := Power[E,Hermitian[P]] /; OperatorQ[P]


(* ::Subsection:: *)
(* Region RULES FOR COMMUTATION *)


 Commutator[P_, 0] := 0 

 Commutator[0, P_] := 0 

 Commutator[P_ ,Operator[0]] := 0

 Commutator[Operator[0], P_] := 0

 Commutator[P_, Operator[1]] := 0 /; OperatorQ[P]

 Commutator[Operator[1], P_] := 0 /; OperatorQ[P]

 Commutator[P_, P_] := 0 /; OperatorQ[P]

 Commutator[P_, n_ Q_] := 
 	n Commutator[P, Q] /; 
	!OperatorQ[n] && OperatorQ[P] && OperatorQ[Q] 
 
 Commutator[m_ P_, Q_] := 
 	m Commutator[P, Q] /; 
	!OperatorQ[m] && OperatorQ[P] && OperatorQ[Q]

 Commutator[P_, Q_+ R_] := 
 	Commutator[P, Q] + Commutator[P, R] /; 
	OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[P_+ Q_, R_] := 
 	Commutator[P, R] + Commutator[Q, R] /; 
	OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[P_, HoldPattern[Q_** R_]] := 
 	Commutator[P, Q] ** R + Q ** Commutator[P, R] /; 
	OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[HoldPattern[P_** Q_], R_] := 
 	Commutator[P, R] ** Q + P ** Commutator[Q, R]/; 
	OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Hermitian[HoldPattern[Commutator[P_, Q_]]] := 
 	Commutator[Hermitian[Q], Hermitian[P]]/; 
	OperatorQ[P] && OperatorQ[Q]

 Commutator[X:(Operator[__] | Operator[__][__]), expr_?OperatorQ] := 0 /; checkOperator[X, expr]

 Commutator[expr_?OperatorQ, X:(Operator[__] | Operator[__][t__])] := 0 /; checkOperator[X, expr] 


(* ::Subsection:: *)
(* Region RULES FOR QASet AND QAClear *)

(* *************** DO NOT CHANGE ORDER - ORDER MATTERS ! ************ *)	


 SetAttributes[ QASet, {HoldAllComplete} ]
 SetAttributes[ QAClear, {HoldAllComplete} ]
 
 		                   
 QASet[Set[Commutator[Commutator[P_, Q_], Commutator[R_, S_]] , T_]] :=
	Module[ { },
		Commutator[Commutator[P, Q], Commutator[R, S]] = T;
		Commutator[Commutator[P, Q], Commutator[S, R]] = -T;
		Commutator[Commutator[Q, P], Commutator[R, S]] = -T;
		Commutator[Commutator[Q, P], Commutator[S, R]] = T;
		Commutator[Commutator[R, S], Commutator[P, Q]] = -T;
		Commutator[Commutator[R, S], Commutator[Q, P]] = T;
		Commutator[Commutator[S, R], Commutator[P, Q]] = T;
		Commutator[Commutator[S, R], Commutator[Q, P]] = -T;
	]

 QASet[Set[
 	( 
 		Commutator[P_, Commutator[Q_, R_]] |
      	Commutator[Commutator[R_, Q_], P_] 
	),  S_ ]] :=
 	Module[{ }, 
		Commutator[P, Commutator[Q, R]] = S;
        Commutator[P, Commutator[R, Q]] = -S;
        Commutator[Commutator[Q, R], P] = -S;
        Commutator[Commutator[R, Q], P] = S; 
    ]

 QASet[Set[Commutator[P_, Q_], 0]] :=
	Module[ {  },
		Commutator[P, Q] = 0;
		Commutator[Q, P] = 0;
		Commutator[P, x_] /; 
			checkOperator[Q, x] :=
			0;
		Commutator[x_, P] /; 
			checkOperator[Q, x] :=
			0;
		Commutator[Q, x_] /; 
			checkOperator[P, x] :=
			0;
		Commutator[x_, Q] /; 
			checkOperator[P, x] :=
			0;
		Commutator[x_, y_] /; 
			checkOperator[P, x] && checkOperator[Q, y] :=
			0;
		Commutator[x_, y_] /; 
			checkOperator[Q, x] && checkOperator[P, y] :=
			0;
	];

 QASet[Set[Commutator[P_, Q_], R_]] :=
	Module[{ k, auxP, auxQ },
		Commutator[P, Q] = R;
		Commutator[Q, P] = -R;
		auxP = (P /. Verbatim[Pattern][x_, _] :> x); 
		auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);                  
		If[P === auxP && Q === auxQ && Commutator[P, R] === Commutator[Q, R] === 0,
			Commutator[P^n_, Q^m_] /; 
				IntegerQ[n] && IntegerQ[m] :=
			 	Sum[Binomial[n, k] (m! / (m - k)!) (R^k) ** Q^(m - k) ** P^(n - k), {k, 1, n}]; 
			Commutator[Q^n_, P^m_] /; 
				IntegerQ[n] && IntegerQ[m] :=
			 	Sum[Binomial[n, k] (m! / (m - k)!) ((-1)^k)(R^k) ** P^(m - k) ** Q^(n - k), {k, 1, n}];
	 	    Commutator[P, x_] /; 
	 	    	checkOperator[Q, x] := 
	 	    	R ** (D[x, Q] /. Times -> NonCommutativeMultiply) // MapAll[QAProductExpand, #]&;
		    Commutator[x_, P] /; 
		    	checkOperator[Q, x] := 
		    	-R ** (D[x, Q] /. Times -> NonCommutativeMultiply) // MapAll[QAProductExpand, #]&;
		    Commutator[Q, y_] /; 
		    	checkOperator[P, y] := 
		    	-R ** (D[y, P] /. Times -> NonCommutativeMultiply) // MapAll[QAProductExpand, #]&;
		    Commutator[y_, Q] /; 
		    	checkOperator[P, y] := 
		    	R ** (D[y, P] /. Times -> NonCommutativeMultiply) // MapAll[QAProductExpand, #]&;
		]; 
	] /; R =!= 0

 QASet[ 
 	Set[ 
 	(
 		Commutator[P_, Commutator[P_, Q_]] |
		Commutator[P_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], P_] |
		Commutator[Commutator[Q_, P_], P_]  
	) , 0 ],
	Set[ 
	( 
		Commutator[Q_, Commutator[P_, Q_]] |
		Commutator[Q_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], Q_] |
		Commutator[Commutator[Q_, P_], Q_]  
	) , 0 ] ] :=
	Module[ { k },
		Commutator[P, Commutator[P, Q]] = 0;
		Commutator[P, Commutator[Q, P]] = 0;
		Commutator[Commutator[P, Q], P] = 0;
		Commutator[Commutator[Q, P], P] = 0;
		Commutator[Q, Commutator[P, Q]] = 0;
		Commutator[Q, Commutator[Q, P]] = 0;
		Commutator[Commutator[P, Q], Q] = 0;
		Commutator[Commutator[Q, P], Q] = 0;
		Commutator[P^n_, Q^m_]  /; 
			IntegerQ[n] && IntegerQ[m] :=
			Sum[Binomial[n, k] (m! / (m - k)!) (Commutator[P,Q]^k) ** Q^(m-k) ** P^(n-k), {k, 1, n}];
		Commutator[Q^n_, P^m_] /; 
			IntegerQ[n] && IntegerQ[m] :=
			Sum[Binomial[n, k] (m! / (m - k)!) ((-1)^k) (Commutator[P, Q]^k) ** P^(m-k) ** Q^(n-k), {k, 1, n}];
		Commutator[P, x_] /; 
			checkOperator[Q, x] :=
			Commutator[P, Q] ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[x_, P] /; 
			checkOperator[Q, x] :=
			-Commutator[P, Q] ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[Q, y_] /; 
			checkOperator[P, y] :=
			-Commutator[P, Q] ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[y_, Q] /; 
			checkOperator[P, y] :=
			Commutator[P, Q] ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
   ]

 QASet[Set[P_?OperatorQ ** Ket[m__], c_. Ket[n__]]] :=
	Module[ {y = P},
		Unprotect[NonCommutativeMultiply];
		P ** Ket[m] = c Ket[n];
		Bra[m] ** Hermitian[P] = Conjugate[c] Bra[n];
		If[ (Ket[m] /. Verbatim[Pattern][x_, _] :> x ) === Ket[n],
			x_ ** Ket[m] /; 
				(OperatorQ[x] && 
				checkOperator[P,x] && 
				Head[x] =!= NonCommutativeMultiply) :=
				QAProductExpand[( x Ket[n] /. y :> c )];
			Bra[m] ** x_ /; 
				(OperatorQ[x] && 
				checkOperator[Hermitian[P],x] &&
				Head[x] =!= NonCommutativeMultiply) :=
				QAProductExpand[( x Bra[n] /. Hermitian[y] :> Conjugate[c])];
     	];
     	Protect[NonCommutativeMultiply];
   	]

 (*	NEW: trying to implement differential operators
 	TODO: handle subscripts / superscripts appropriately
 *)
 QASet[Set[Bra[m__] ** P_?OperatorQ ** Ket[n__], Diff_]] := 
 	Module[{m1, n1},
 		m1 = Sequence @@ (First /@ { m });
 		n1 = Sequence @@ (First /@ { n });
		Unprotect[NonCommutativeMultiply];
		Bra[m] ** P ** Ket[n] = Diff[BraKet[{m1}, {n1}], m1];
     	Protect[NonCommutativeMultiply];
 	]


 QASet[SetDelayed[Commutator[Commutator[P_, Q_], Commutator[R_, S_]], T_]] :=
	Module[ { },
		Commutator[Commutator[P, Q], Commutator[R, S]] := T;
		Commutator[Commutator[P, Q], Commutator[S, R]] := -T;
		Commutator[Commutator[Q, P], Commutator[R, S]] := -T;
		Commutator[Commutator[Q, P], Commutator[S, R]] := T;
		Commutator[Commutator[R, S], Commutator[P, Q]] := -T;
		Commutator[Commutator[R, S], Commutator[Q, P]] := T;
		Commutator[Commutator[S, R], Commutator[P, Q]] := T;
		Commutator[Commutator[S, R], Commutator[Q, P]] := -T;
	]

 QASet[SetDelayed[
 	( 
 		Commutator[P_, Commutator[Q_, R_]] |
		Commutator[Commutator[R_, Q_], P_] 
	), S_]] :=
	Module[ { },
		Commutator[P, Commutator[Q, R]] := S;
		Commutator[P, Commutator[R, Q]] := -S;
		Commutator[Commutator[Q, R], P] := -S;
		Commutator[Commutator[R, Q], P] := S;
   ]

 QASet[SetDelayed[Commutator[P_, Q_], R_]] :=
	Module[ { k, auxP, auxQ},
		Commutator[P, Q] := R;
		Commutator[Q, P] := -R;
		auxP = (P /. Verbatim[Pattern][x_, _] :> x);
		auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);
		If[ P === auxP && Q === auxQ && Commutator[P, R] === Commutator[Q, R] === 0,
			Commutator[P^n_, Q^m_] /; 
				IntegerQ[n] && IntegerQ[m] :=
				Sum[Binomial[n,k] (m!/(m - k)!) (R^k) ** Q^(m-k) ** P^(n-k), {k, 1, n}];
			Commutator[Q^n_, P^m_] /; 
				IntegerQ[n] && IntegerQ[m] :=
				Sum[Binomial[n,k] (m!/(m - k)!) ((-1)^k) (R^k) ** P^(m-k) ** Q^(n-k), {k, 1, n}];
			Commutator[P, x_] /; 
				checkOperator[Q, x] :=
				R ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
			Commutator[x_, P] /; 
				checkOperator[Q, x] :=
				-R ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
			Commutator[Q, y_] /; 
				checkOperator[P, y] :=
				-R ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
			Commutator[y_, Q] /; 
				checkOperator[P, y] :=
				R ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		];
	] /; R =!= 0

 QASet[SetDelayed[Commutator[P_, Q_], 0]] :=
	Module[ {  },                  
		Commutator[P, Q] := 0;
		Commutator[Q, P] := 0;
		Commutator[P, x_] /; checkOperator[Q, x] := 0;
		Commutator[x_, P] /; checkOperator[Q, x] := 0;
		Commutator[Q, x_] /; checkOperator[P, x] := 0;
		Commutator[x_, Q] /; checkOperator[P, x] := 0; 
		Commutator[x_, y_] /; checkOperator[P, x] && checkOperator[Q, y] := 0;
		Commutator[x_, y_] /; checkOperator[Q, x] && checkOperator[P, y] := 0; 
	]

 (*	TODO: no reason why 1 SetDelayed and 1 Set -> changed Set -> SetDelayed
 *)
 QASet[
 	SetDelayed[ 
 	( 
 		Commutator[P_, Commutator[P_, Q_]] |
		Commutator[P_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], P_] |
		Commutator[Commutator[Q_, P_], P_]  
	), 0 ],
	SetDelayed[ 
	( 
		Commutator[Q_, Commutator[P_, Q_]] |
		Commutator[Q_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], Q_] |
		Commutator[Commutator[Q_, P_], Q_]  
	), 0 ]] :=
	Module[ { k },	
		Commutator[P, Commutator[P, Q]] := 0;
		Commutator[P, Commutator[Q, P]] := 0;
		Commutator[Commutator[P, Q], P] := 0;
		Commutator[Commutator[Q, P], P] := 0;
		Commutator[Q, Commutator[P, Q]] := 0;
		Commutator[Q, Commutator[Q, P]] := 0;
		Commutator[Commutator[P, Q], Q] := 0;
		Commutator[Commutator[Q, P], Q] := 0;
		Commutator[P^n_, Q^m_]  /; 
			IntegerQ[n] && IntegerQ[m] := 
			Sum[Binomial[n,k] (m!/(m-k)!) (Commutator[P,Q]^k) ** Q^(m-k) ** P^(n-k), {k, 1, n}];
		Commutator[Q^n_, P^m_] /; 
			IntegerQ[n] && IntegerQ[m] :=
			Sum[Binomial[n,k] (m! / (m - k)!) ((-1)^k) (Commutator[P,Q]^k) ** P^(m-k) ** Q^(n-k), {k, 1, n}];
		Commutator[P, x_] /; 
			checkOperator[Q, x] := 
			Commutator[P, Q] ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[x_, P] /; 
			checkOperator[Q, x] := 
			-Commutator[P, Q] ** (D[x, Q] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[Q, y_] /; 
			checkOperator[P, y] := 
			-Commutator[P, Q] ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&;
		Commutator[y_, Q] /; 
			checkOperator[P, y] := 
			Commutator[P, Q] ** (D[y, P] /. Times->NonCommutativeMultiply) // MapAll[QAProductExpand,#]&; 
	]

 QASet[SetDelayed[P_?OperatorQ ** Ket[m__], c_. Ket[n__]]] :=
	Module[ {y = P}, 
		Unprotect[NonCommutativeMultiply];
	   	P ** Ket[m] := c Ket[n];
		Bra[m] ** Hermitian[P] := Conjugate[c] Bra[n];
		If[ (Ket[m] /. Verbatim[Pattern][x_, _] :> x ) === Ket[n],
			x_ ** Ket[m] /; 
				(OperatorQ[x] && 
				checkOperator[P, x] && 
				Head[x] =!= NonCommutativeMultiply) := 
				QAProductExpand[( x Ket[n] /. y :> c )];
			Bra[m] ** x_ /; 
				(OperatorQ[x] && 
				checkOperator[Hermitian[P], x] &&
				Head[x] =!= NonCommutativeMultiply) := 
				QAProductExpand[( x Bra[n] /. Hermitian[y] :> Conjugate[c])]; 
		]; 
		Protect[NonCommutativeMultiply]; 
	]

 QASet[SetDelayed[Bra[m__] ** P_?OperatorQ ** Ket[n__], Diff_]] :=
 	Module[{m1, n1},
 		m1 = Sequence @@ (First /@ { m });
 		n1 = Sequence @@ (First /@ { n });
		Unprotect[NonCommutativeMultiply];
		Bra[m] ** P ** Ket[n] := Diff[BraKet[{m1}, {n1}], m1];
     	Protect[NonCommutativeMultiply];
 	]


 QAClear[Unset[Commutator[Commutator[P_, Q_], Commutator[R_, S_]]]] :=
	Module[ { },
		Commutator[Commutator[P, Q],Commutator[R, S]] =.;
		Commutator[Commutator[P, Q],Commutator[S, R]] =.;
		Commutator[Commutator[Q, P],Commutator[R, S]] =.;
		Commutator[Commutator[Q, P],Commutator[S, R]] =.;
		Commutator[Commutator[R, S],Commutator[P, Q]] =.;
		Commutator[Commutator[R, S],Commutator[Q, P]] =.;
		Commutator[Commutator[S, R],Commutator[P, Q]] =.;
		Commutator[Commutator[S, R],Commutator[Q, P]] =.; 
	]

 QAClear[Unset[ 
 	( 
 		Commutator[P_, Commutator[Q_,R_]] |
		Commutator[Commutator[R_,Q_], P_] 
	)]] :=
	Module[ { }, 
		Commutator[P,Commutator[Q,R]] =.;
		Commutator[P,Commutator[R,Q]] =.;
		Commutator[Commutator[Q,R],P] =.;
		Commutator[Commutator[R,Q],P] =.; 
	]

 QAClear[Unset[Commutator[P_, Q_]]] :=
	Module[ { auxP, auxQ, T = Commutator[P, Q],
			  R = Commutator[P, Commutator[P, Q]], 
			  S = Commutator[Q, Commutator[P, Q]]},
		Commutator[P, Q] =.;
		Commutator[Q, P] =.;
		auxP = (P /. Verbatim[Pattern][x_, _] :> x); 
		auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);                  
		If[P === auxP && Q === auxQ && R === S === 0 && T =!= 0,
			Commutator[P^n_, Q^m_] /; IntegerQ[n] && IntegerQ[m] =.;
			Commutator[Q^n_, P^m_] /; IntegerQ[n] && IntegerQ[m] =.;
			Commutator[P, x_] /; checkOperator[Q, x] =.;
			Commutator[x_, P] /; checkOperator[Q, x] =.;
			Commutator[Q, y_] /; checkOperator[P, y] =.;
			Commutator[y_, Q] /; checkOperator[P, y] =.; 
		];
		If[T===0,
			Commutator[P, x_] /; checkOperator[Q, x] =.;
			Commutator[x_, P] /; checkOperator[Q, x] =.;
			Commutator[Q, x_] /; checkOperator[P, x] =.;
			Commutator[x_, Q] /; checkOperator[P, x] =.;
			Commutator[x_, y_] /; checkOperator[P, x] && checkOperator[Q, y] =.;
			Commutator[x_, y_] /; checkOperator[Q, x] && checkOperator[P, y] =.;
		]; 
	]

 QAClear[ 
 	Unset[ 
 	( 
 		Commutator[P_, Commutator[P_, Q_]] |
		Commutator[P_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], P_] |
		Commutator[Commutator[Q_, P_], P_]  
	)],
	Unset[ 
	( 
		Commutator[Q_, Commutator[P_, Q_]] |
		Commutator[Q_, Commutator[Q_, P_]] |
		Commutator[Commutator[P_, Q_], Q_] |
		Commutator[Commutator[Q_, P_], Q_]  
	)]] :=
	Module[ {  },	
		Commutator[P, Commutator[P, Q]] =.;
		Commutator[P, Commutator[Q, P]] =.;
		Commutator[Commutator[P, Q], P] =.;
		Commutator[Commutator[Q, P], P] =.;
		Commutator[Q, Commutator[P, Q]] =.;
		Commutator[Q, Commutator[Q, P]] =.;
		Commutator[Commutator[P, Q], Q] =.;
		Commutator[Commutator[Q, P], Q] =.;
		Commutator[P^n_, Q^m_] /; IntegerQ[n] && IntegerQ[m] =.;
		Commutator[Q^n_, P^m_] /; IntegerQ[n] && IntegerQ[m] =.;
		Commutator[P, x_] /; checkOperator[Q, x] =.;
		Commutator[x_, P] /; checkOperator[Q, x] =.;
		Commutator[Q, y_] /; checkOperator[P, y] =.;
		Commutator[y_, Q] /; checkOperator[P, y] =.; 
	]

 QAClear[Unset[P_?OperatorQ ** Ket[m__]]] :=
	Module[ { }, 
		Unprotect[NonCommutativeMultiply];
	   	P ** Ket[m] =.;
		Bra[m] ** Hermitian[P] =.;
		Off[Unset::norep];
		x_ ** Ket[m] /; 
			(OperatorQ[x] && 
			checkOperator[P, x] && 
			Head[x] =!= NonCommutativeMultiply) =.;
		Bra[m] ** x_ /; 
			(OperatorQ[x] && 
			checkOperator[Hermitian[P], x] &&
			Head[x] =!= NonCommutativeMultiply) =.;
		On[Unset::norep]; 
		Protect[NonCommutativeMultiply]; 
	] 

 QAClear[Unset[Bra[m__] ** P_?OperatorQ ** Ket[n__]]] :=
 	Module[{},
		Unprotect[NonCommutativeMultiply];
		Bra[m] ** P ** Ket[n] = .;
     	Protect[NonCommutativeMultiply];
 	]


(* ::Subsection:: *)
(* Region RULES FOR EXPANDING SERIES *)

 SetAttributes[QASeries, {HoldAllComplete}]

 QASeries[expr_, {x_, a_, n_}] := 
 	Module[{k, t, f},
		t = (((expr /. x -> a) /. Operator[1] -> 1) Operator[1] +
			Sum[(( (D[expr, {x, k}] /. x -> a) /. Operator[1] -> 1) (x - a)^k) / f[k], {k, 1, n}]) 
			/. Operator[0] -> 0;
		t = t /. f[v_] :> HoldForm[v!]
	]

 QASeries[c_. Exp[P_] ** R_ ** Exp[Q_], n_] :=
	Module[{f, k, s},
		s = c Sum[Nest[Commutator[P, #]&, R, k] / f[k], {k, 0, n-1}];
		s = s /. f[v_] :> HoldForm[v!]
	] /; 
	OperatorQ[P] && OperatorQ[R] && P === -Q && Commutator[P, R] =!= 0


(* ::Subsection:: *)
(* Region RULES FOR SIMPLIFYING QUANTUM OBJECTS *)


 QAOrdering[expr_, list_List] := expr //. CommutationOrderRules[list]

 QAPowerExpand[expr_] := expr //. rule1

 QAPowerContract[expr_] := expr //. rule2

 QAProductExpand[expr_?(!SimpleConstantQ[#, "10."]&)] := expr //. rule4
 
 QAProductExpand[expr_?(SimpleConstantQ[#, "11."]&)] := expr //. NonCommutativeMultiply -> Times
 
 QAProductExpand[expr_] := expr

 QAExpandAll[expr_?OperatorQ] := 
 	expr // QAPowerExpand // QAProductExpand // QACommutatorExpand // Expand

 QAExpandAll[expr_] := expr

 QACommutatorExpand[expr_?OperatorQ] := expr //. rule3 
 
 QACommutatorExpand[expr_] := expr

 QACommutatorContract[expr_, t : {{_, _} ...}] := 
	Fold[ReplaceRepeated, expr,                        			
		Map[First[#] ** Last[#] :>
			Commutator[First[#], Last[#]] + Last[#] ** First[#] &, 
			t]
		]

 QACommutatorContract[expr_, {P_, Q_}] := expr //. P ** Q :> Commutator[P, Q] + Q ** P

 QACommutatorContract[expr_] := expr //. P_ ** Q_ - Q_ ** P_ :> Commutator[P, Q]


(* ::Subsection:: *)
(* Region Auxiliar Definitions and Rules *)
 
 (* Brings Operators of Commutators in given order 
 *)
 CommutationOrderRules[l_List] := 
	Module[
		{x = {First[l]}, y = Rest[l], f, sol = {}},
		f = Rule[#2**#1, #1**#2 + Commutator[#2, #1]] &;
		While[Length[y] >= 1, 
			AppendTo[sol, Outer[f, x, y]];
			x = {First[y]};
			y = Rest[y]
		];
		Flatten[sol]
	]

(*	Evaluates a NonCommutativeMultiply expression
	- Constants and Quantum objects are separated
	- Quantum objects are multiplied non commutatively
	- Constants are multiplied commutatively with each other and the Quantum objects
*)
 expand[NonCommutativeMultiply[P__]] :=
	Module[{temp1, temp2, l1,  temp3 = {P}},
		temp1 = Select[temp3, (!SimpleConstantQ[#, "12."]&)];	
		temp2 = Select[temp3, (SimpleConstantQ[#, "13."]&)];
		l1 = Length[temp1];
		Which[
			l1 > 1, 
			Times[Apply[NonCommutativeMultiply, temp1], Apply[Times, temp2]],
			l1 === 1, 
			First[temp1] Apply[Times, temp2],
			l1 === 0, 
			Apply[Times, temp2]
		]
	]

 (* Power law for 
 	- power of sum
 	- power of Operator 
 *)
 rule1 = 
 	{
 		(P_Plus)^n_?IntegerQ /; 
			OperatorQ[P] && n > 0 :> 
			Plus @@ Flatten[Outer[NonCommutativeMultiply, Sequence @@ Table[List @@ P, {n}]]],
		Power[P_?OperatorQ, n_?IntegerQ] /; 
			n > 0 :>
			Apply[NonCommutativeMultiply, Table[P, {n}]],
		Power[P_?OperatorQ, n_?IntegerQ] /; 
			n < -1 :>
			Apply[NonCommutativeMultiply, Table[P^(-1), {-n}]]
	}

 (* Power law for product of powers *)
 rule2 = P_^n_. ** P_^m_. :> P^(n + m) /; OperatorQ[P]

 (* Commutator expansion *)
 rule3 = HoldPattern[Commutator[P_, Q_]] :> P ** Q - Q ** P

 (* Expansion of NonCommutativeMultiply *)
 rule4 = NonCommutativeMultiply[P__] :> expand[NonCommutativeMultiply[P]]

 (* Performs a check that:
 	- the given Operator is the only one in expr
 	- is exactly the same Operator contained
 	- is not the same as expr itself
 *)
 checkOperator[Operator[P__],expr_] :=
	Complement[
		Cases[expr, (Operator[__]|Operator[__][__]), {0, Infinity}], 
		{Operator[0],Operator[1]} 
	] ===
	Union[
		Cases[expr, Operator[P], {0, Infinity}]
	] && expr =!= Operator[P]

 checkOperator[Operator[P__][t__],expr_] :=
	Complement[
		Cases[expr, (Operator[__]|Operator[__][__]), {0, Infinity}],
		{Operator[0],Operator[1]} 
	] ===
	Union[
		Cases[expr, Operator[P][t], {0, Infinity}]
	] && expr =!= Operator[P][t]


Protect[ Release[protected] ]


End[]

EndPackage[]

